import numpy as np
import h5py
import os

import keras
from keras.models import Sequential
from keras.layers import LSTM, Dense, Bidirectional, Dropout, merge, BatchNormalization, GRU, TimeDistributed
from keras.layers.core import *
from keras.layers.recurrent import LSTM
from keras.models import *

import sklearn.metrics as sklm

import pandas as pd

use_pitch = False
num_phases_to_predict = 1  # 1=stroke/noStroke, 4=prep/stroke/hold/other, 6=prep/stroke/hold/partialRetract/retract/none

data_directory = '/your/data/directory/'
model_directory = '/your/model/directory/'
# replace file_takes with your file naming
file_takes = ['01', '02', '04', '05', '07', '08', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '21', '22', '23', '25', '26', '27', '30']
# motion format:
# - 20 fps
# - joints: shoulder, elbow, wrist, and each fingertip (total of 16 joints, 8 for each side)
# - motion representation: per-frame x, y, and z offset (total of 48 values (3 for each of the 16 joints))
# (the offset is the current frame's positional value minus the previous frame's positional value)
# - z-score normalized (MATLAB: "normalize")
#
# motion file should be in csv format, with the columns in the following manner:
# left_shoulder_x,  left_shoulder_y,  left_shoulder_z,  left_elbow_x,  left_elbow_y,  left_elbow_z, ..

# pitch format, if using:
# - 20 fps
# - single column, z-score normalized

######################## OUTPUT CONVENTION:
# for 1 phase:
# 0 = no stroke
# 1 = stroke

# for 4 phases:
# 0 = hold
# 1 = 'other'
# 2= 'prep'
# 3 = 'stroke'

# for 6 phases:
# 0 = hold
# 1 = none
# 2 = partial retract
# 3 = prep
# 4 = retract
# 5 = stroke

####################################### MODELS
timesteps = 100
data_dim = 48

def make_classifier_ph1_pitch():
    dropout = 0.5
    rec_dropout = 0.5
    h = 256
    h_dense = 128
    dense_drop = 0.3
    drop_out_1 = 0.5
    drop_out_2 = 0.5
    inputs = Input(shape=(timesteps, data_dim,))
    pitch_input = Input(shape=(timesteps, 1,))
    in_with_pitch = keras.layers.concatenate([inputs, pitch_input])
    dense_input = TimeDistributed(Dense(h_dense))(in_with_pitch)
    drop_1 = Dropout(dense_drop)(dense_input)
    bi_layer_1 = Bidirectional(LSTM(h, dropout=dropout, recurrent_dropout=rec_dropout, return_sequences=True))(drop_1)
    batch_norm = BatchNormalization(axis=2)(bi_layer_1)
    output = Dense(num_phases_to_predict, activation='sigmoid', name='out_softmax')(batch_norm)
    model = Model(inputs=[inputs, pitch_input], outputs=output)
    return model


def make_classifier_ph1_noPitch():
    dropout = 0.5
    rec_dropout = 0.5
    h = 256
    h_dense = 128
    dense_drop = 0.3
    drop_out_1 = 0.5
    drop_out_2 = 0.5
    inputs = Input(shape=(timesteps, data_dim,))
    dense_input = TimeDistributed(Dense(h_dense))(inputs)
    drop_1 = Dropout(dense_drop)(dense_input)
    bi_layer_1 = Bidirectional(LSTM(h, dropout=dropout, recurrent_dropout=rec_dropout, return_sequences=True))(drop_1)
    batch_norm = BatchNormalization(axis=2)(bi_layer_1)
    output = Dense(num_phases_to_predict, activation='sigmoid', name='out_softmax')(batch_norm)
    model = Model(inputs=[inputs], outputs=output)
    return model

def make_classifier_ph4_noPitch():
    h = 128
    h_dense = 128
    inputs = Input(shape=(timesteps, data_dim,))
    dense_input = TimeDistributed(Dense(h_dense))(inputs)
    drop_1 = Dropout(0)(dense_input)
    layer_1 = LSTM(h, return_sequences=True)(drop_1)
    layer_2 = Bidirectional(LSTM(h, return_sequences=True))(layer_1)
    batch_norm = BatchNormalization(axis=2)(layer_2)
    output = Dense(num_phases_to_predict, activation='softmax', name='out_softmax')(batch_norm)
    model = Model(inputs=[inputs], outputs=output)
    return model

def make_classifier_ph4_pitch():
    h = 128
    h_dense = 128
    inputs = Input(shape=(timesteps, data_dim,))
    pitch_input = Input(shape=(timesteps, 1,))
    in_with_pitch = keras.layers.concatenate([inputs, pitch_input])
    dense_input = TimeDistributed(Dense(h_dense))(in_with_pitch)
    drop_1 = Dropout(0)(dense_input)
    layer_1 = LSTM(h, return_sequences=True)(drop_1)
    layer_2 = Bidirectional(LSTM(h, return_sequences=True))(layer_1)
    batch_norm = BatchNormalization(axis=2)(layer_2)
    output = Dense(num_phases_to_predict, activation='softmax', name='out_softmax')(batch_norm)
    model = Model(inputs=[inputs, pitch_input], outputs=output)
    return model

def make_classifier_ph6_pitch():
    dropout = 0.3
    h = 256
    h_dense = 128
    dense_drop = 0.3
    drop_out_1 = 0.5
    drop_out_2 = 0.5
    inputs = Input(shape=(timesteps, data_dim,))
    pitch_input = Input(shape=(timesteps, 1,))
    in_with_pitch = keras.layers.concatenate([inputs, pitch_input])
    dense_input = TimeDistributed(Dense(h_dense))(in_with_pitch)
    drop_1 = Dropout(dense_drop)(dense_input)
    layer_1 = LSTM(h, dropout=dropout, return_sequences=True)(drop_1)
    drop_1 = Dropout(drop_out_1)(layer_1)
    layer_2 = Bidirectional(LSTM(h, dropout=dropout, return_sequences=True))(drop_1)
    drop_2 = Dropout(drop_out_2)(layer_2)
    batch_norm = BatchNormalization(axis=2)(drop_2)
    output = Dense(num_phases_to_predict, activation='softmax', name='out_softmax')(batch_norm)
    model = Model(inputs=[inputs,pitch_input], outputs=output)
    return model

def make_classifier_ph6_noPitch():
    dropout = 0.3
    h = 256
    h_dense = 128
    dense_drop = 0.3
    drop_out_1 = 0.5
    drop_out_2 = 0.5
    inputs = Input(shape=(timesteps, data_dim,))
    dense_input = TimeDistributed(Dense(h_dense))(inputs)
    drop_1 = Dropout(dense_drop)(dense_input)
    layer_1 = LSTM(h, dropout=dropout, return_sequences=True)(drop_1)
    drop_1 = Dropout(drop_out_1)(layer_1)
    layer_2 = Bidirectional(LSTM(h, dropout=dropout, return_sequences=True))(drop_1)
    drop_2 = Dropout(drop_out_2)(layer_2)
    batch_norm = BatchNormalization(axis=2)(drop_2)
    output = Dense(num_phases_to_predict, activation='softmax', name='out_softmax')(batch_norm)
    model = Model(inputs=[inputs], outputs=output)
    return model

if num_phases_to_predict == 1:
    if use_pitch:
        model_num = "ph1_pitch"
        model = make_classifier_ph1_pitch()
    else:
        model_num = "ph1_noPitch"
        model = make_classifier_ph1_noPitch()
elif num_phases_to_predict == 4:
    if use_pitch:
        model_num = "ph4_pitch"
        model = make_classifier_ph4_pitch()
    else:
        model_num = "ph4_noPitch"
        model = make_classifier_ph4_noPitch()
elif num_phases_to_predict == 6:
    if use_pitch:
        model_num = "ph6_pitch"
        model = make_classifier_ph6_pitch()
    else:
        model_num = "ph6_noPitch"
        model = make_classifier_ph6_noPitch()
else:
    raise ValueError

####################################### READ DATA

def readData(file_take):
    motion = pd.read_csv(data_directory + file_take + '_motion_normed.csv', sep=',', header=None)
    pitch = pd.read_csv(data_directory + file_take + '_pitch_normed.csv', sep=',', header=None)
    return motion, pitch

def readData_noPitch(file_take):
    motion = pd.read_csv(data_directory + file_take + '_motion_normed.csv', sep=',', header=None)
    return motion

# model.summary()
model.load_weights(model_directory + model_num + '.hdf5')

for file_take in file_takes:
    if use_pitch:
        [x_train, pitch] = readData(file_take)
		pitch = np.array(pitch)
	else:
        x_train = readData_noPitch(file_take)
    x_train = np.array(x_train)
    if x_train.shape[0] < x_train.shape[1]:
        x_train = np.transpose(x_train)
    print(x_train.shape)

    num_samples = x_train.shape[0]
    print('num_samples :', num_samples)
    cut_off = num_samples - (int(num_samples / timesteps) * timesteps)
    
	if x_train.shape[1]==data_dim*2:
	    has_positions = True
	else:
		has_positions = False
	
    if cut_off > 0:
        if has_positions:
	        x_train = x_train[:-cut_off, 1::2]
		else:
            x_train = x_train[:-cut_off, :]
		if use_pitch:
            pitch = pitch[:-cut_off]
    else:
	    if has_positions:
            x_train = x_train[:, 1::2]

    #if num_phases_to_predict == 6 and not predict_nt:
    #    x_train = np.delete(x_train, [0, 1, 2, 27, 28, 29], axis=1)

    num_samples = x_train.shape[0]
    print(x_train.shape)
    print('num_samples :', num_samples)
    x_train = x_train.reshape((int(num_samples / timesteps), timesteps, data_dim))
    if use_pitch:
	    pitch = pitch.reshape((int(num_samples / timesteps), timesteps, 1))
    print('X shape (num_samples, timesteps, data_dim):', x_train.shape)

    ##### PREDICT
    if use_pitch:
        y_pred = model.predict([x_train, pitch])
    else:
        y_pred = model.predict([x_train])

    if num_phases_to_predict > 1:
        preds = np.argmax(y_pred, axis=2)
    else:
        preds = np.round(y_pred)

    preds = preds.reshape((preds.shape[0] * timesteps))

    ##### PRINT TO FILE
    np.savetxt(data_directory + "prediction_" + file_take + ".csv", preds, delimiter=",")


